#****************************************************************************
#**
#**  File     :  /lua/unit.lua
#**  Author(s):  John Comes, David Tomandl, Gordon Duclos
#**
#**  Summary  : The Unit lua module
#**
#**  Copyright  2005 Gas Powered Games, Inc.  All rights reserved.
#****************************************************************************

local Entity = import('/lua/sim/Entity.lua').Entity
local explosion = import('/lua/defaultexplosions.lua')
local EffectTemplate = import('/lua/EffectTemplates.lua')
local EffectUtilities = import('/lua/EffectUtilities.lua')
local Game = import('/lua/game.lua')
local utilities = import('/lua/utilities.lua')
local Shield = import('/lua/shield.lua').Shield
local UnitShield = import('/lua/shield.lua').UnitShield
local AntiArtilleryShield = import('/lua/shield.lua').AntiArtilleryShield
local Buff = import('/lua/sim/buff.lua')
local AIUtils = import('/lua/ai/aiutilities.lua')

local oldUnit=Unit
Unit = Class(oldUnit) {
	OnCollisionCheck = function(self, other, firingWeapon)
      if other.lastimpact and other.lastimpact == self:GetEntityId() then return false end 
        return oldUnit.OnCollisionCheck(self, other, firingWeapon)
    end,
    
	-- This thread runs constantly in the background for all units. It ensures that the cloak effect and cloak field are always in the correct state
	CloakEffectControlThread = function(self)
		local bp = self:GetBlueprint()
		local bpDisplay = bp.Display
		while not (self == nil or self:GetHealth() <= 0 or self:IsDead()) do
			WaitSeconds(0.1)
			self:UpdateCloakEffect()
			local CloakFieldIsActive = self:IsIntelEnabled('CloakField')
			if CloakFieldIsActive then
				local position = self:GetPosition(0)
				-- Range must be (radius - 2) because it seems GPG did that for the actual field for some reason.
				-- Anything beyond (radius - 2) is not cloaked by the cloak field
				local range = bp.Intel.CloakFieldRadius - 2
				local brain = self:GetAIBrain()
				local UnitsInRange = brain:GetUnitsAroundPoint( categories.ALLUNITS, position, range, 'Ally' )
				for num, unit in UnitsInRange do
					unit:MarkUnitAsInCloakField()
				end
			end
		end
	end,

	-- Fork the thread that will deactivate the cloak effect, killing any previous threads that may be running
	MarkUnitAsInCloakField = function(self)
		self.InCloakField = true
		if self.InCloakFieldThread then
			KillThread(self.InCloakFieldThread)
			self.InCloakFieldThread = nil
		end
		self.InCloakFieldThread = self:ForkThread(self.InCloakFieldWatchThread)
	end,

	-- Will deactive the cloak effect if it is not renewed by the cloak field
	InCloakFieldWatchThread = function(self)
		WaitSeconds(0.2)
		self.InCloakField = false
	end,

	-- This is the core of the entire mod. The effect is actually applied here.
	UpdateCloakEffect = function(self)
		local bp = self:GetBlueprint()
		local bpDisplay = bp.Display
		local cloaked = self:IsIntelEnabled('Cloak') or self.InCloakField
		if (not cloaked and self.CloakEffectEnabled) or self:GetHealth() <= 0 then
			self:SetMesh(bpDisplay.MeshBlueprint, true)
		elseif cloaked and not self.CloakEffectEnabled then
			self:SetMesh(bpDisplay.CloakMeshBlueprint , true)
			self.CloakEffectEnabled = true
		end
	end,

	OnStopBeingBuilt = function(self,builder,layer)
		oldUnit.OnStopBeingBuilt(self,builder,layer)
		self:ForkThread(self.CloakEffectControlThread)
	end,

	-- Overridden Function:
	-- Overrode this so that there will be no doubt if the cloak effect is active or not
	SetMesh = function(self, meshBp, keepActor)
		oldUnit.SetMesh(self, meshBp, keepActor)
		self.CloakEffectEnabled = false;
	end,

	-- Overridden Function:
	-- While the CloakEffectControlThread will activate the cloak effect eventually,
	-- this method tries to provide faster a response time to intel changes
	OnIntelEnabled = function(self)
		oldUnit.OnIntelEnabled(self)
		self:UpdateCloakEffect()
	end,

	-- Overridden Function:
	-- While the CloakEffectControlThread will deactivate the cloak effect eventually,
	-- this method tries to provide faster a response time to intel changes
	OnIntelDisabled = function(self)
		oldUnit.OnIntelDisabled(self)
		self:UpdateCloakEffect()
	end,

	InitiateTeleportThread = function(self, teleporter, location, orientation)
	local exbp = self:GetBlueprint()
	local MaxTeleRange = exbp.Defense.MaxTeleRange
	local exdist = VDist3(location, self:GetPosition())
	if MaxTeleRange then
		if exdist > MaxTeleRange then
			LOG('Stopping teleport thread')
			return
		end
	end	
   local brain = self:GetAIBrain()
   local armyIndex = self:GetArmy()
   local mapSizeX, mapSizeZ = GetMapSize()
   local units = GetUnitsInRect(Rect(0,0, mapSizeX, mapSizeZ))
   local filteredUnitList = EntityCategoryFilterDown(categories.ANTITELEPORT, units)
   # - or -
   local unitList = brain:GetUnitsAroundPoint(categories.ANTITELEPORT, location, mapSizeX, 'Enemy')
   for i, unit in filteredUnitList do
      local otherArmyIndex = unit:GetArmy()
      if IsEnemy(armyIndex, otherArmyIndex) then # not needed if using GetUnitsAroundPoint
         local bp = unit:GetBlueprint()
         local noTeleDistance = bp.Defense.NoTeleDistance
         if noTeleDistance then
            local dist = VDist3(location, unit:GetPosition())
            LOG('Anti teleport unit found, dist: ', dist)
            if dist <= noTeleDistance then
               LOG('Stopping teleport thread')
               return
             end
          end
       end
    end
    oldUnit.InitiateTeleportThread(self, teleporter, location, orientation) 
end

}